﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BryanKellyCapstone2017
{
    class AITestFunctions
    {
        public event EventHandler errorTryingToMoveRobot;
        public event EventHandler robotReachedDestination;
        public event EventHandler movingOnVisitedSpace;
        public event EventHandler DFSContinue;
        public event EventHandler DFSReverse;

        private FloorMap floorMap;
        private List<string[]> stateArrayRobot;
        private int depthOfSearch;

        private const string EMPTY = "0";
        private const string WALL = "1";
        private const string ROBOT = "2";
        private const string TARGET = "3";
        private const string GOAL = "G";
        private const string VISITED = "V";

        private int robotRow;
        private int robotColumn;

        
        public void depthFirstSearchRobot(FloorMap  tempFloorMap,int goalRow,int goalColumn)
        {
            floorMap = tempFloorMap;
            depthOfSearch = 0;
            robotRow = -1;
            robotColumn = -1;
            stateArrayRobot = new List<string[]>();
            for (int i = 0; i < floorMap.NumRows; i++)
            {
                stateArrayRobot.Add(new string[floorMap.FloorLayout[i].Length]);
                for(int j = 0; j < floorMap.FloorLayout[i].Length; j++)
                {
                    stateArrayRobot[i][j] = floorMap.FloorLayout[i][j];
                }
            }
            stateArrayRobot[floorMap.RobotRow][floorMap.RobotColumn] = VISITED;
            stateArrayRobot[goalRow][goalColumn] = GOAL;
            if (progressDFSRobot())
            {
                eventRobotReacheDestination();
            }
            else
            {
                eventErrorTryingToRemoveRobot();
            }
            

        }
        private void sendVisitedMessage()
        {
            if (this.movingOnVisitedSpace != null)
                this.movingOnVisitedSpace(this, EventArgs.Empty);
        }
        private void sendDFSContinueMessage()
        {
            if (this.DFSContinue != null)
                this.DFSContinue(this, EventArgs.Empty);
        }
        private void sendDFSReverse()
        {
            if (this.DFSReverse != null)
                this.DFSReverse(this, EventArgs.Empty);
        }
        public bool progressDFSRobot()
        {
            depthOfSearch++;
            if (depthOfSearch > 100)
            {
                return false;
            }
            System.Threading.Thread.Sleep(50);
            string g = stateArrayRobot[floorMap.RobotRow][floorMap.RobotColumn];
            if (stateArrayRobot[floorMap.RobotRow][floorMap.RobotColumn] == GOAL)
                return true;

            
            else
            {
                stateArrayRobot[floorMap.RobotRow][floorMap.RobotColumn] = VISITED;
                g = stateArrayRobot[floorMap.RobotRow][floorMap.RobotColumn];
                if (floorMap.isValidLocation(floorMap.RobotRow - 1, floorMap.RobotColumn) && isEmptyOrGoalSpace(floorMap.RobotRow - 1, floorMap.RobotColumn))
                {
                    if (floorMap.moveRobotNorth(false)) {
                        if (progressDFSRobot())
                            return true;
                        else
                            floorMap.moveRobotSouth();
                    }
                }
                
                if(floorMap.isValidLocation(floorMap.RobotRow, floorMap.RobotColumn-1) && isEmptyOrGoalSpace(floorMap.RobotRow, floorMap.RobotColumn-1))
                {
                    if (floorMap.moveRobotWest(false))
                    {
                        if (progressDFSRobot())
                            return true;
                        else
                            floorMap.moveRobotEast();
                    }
                }
                if (floorMap.isValidLocation(floorMap.RobotRow+1, floorMap.RobotColumn) && isEmptyOrGoalSpace(floorMap.RobotRow + 1, floorMap.RobotColumn))
                {
                    if (floorMap.moveRobotSouth(false))
                    {
                        if (progressDFSRobot())
                            return true;
                        else
                            floorMap.moveRobotNorth();
                    }
                }
                if (floorMap.isValidLocation(floorMap.RobotRow, floorMap.RobotColumn+1) && isEmptyOrGoalSpace(floorMap.RobotRow, floorMap.RobotColumn+1))
                {
                    if(floorMap.moveRobotEast(false))
                    {
                        if (progressDFSRobot())
                            return true;
                        else
                            floorMap.moveRobotWest();
                    }
                }
                return false;
            }
        }
        private bool isEmptyOrGoalSpace(int row,int col)
        {
            return (stateArrayRobot[row][col] == EMPTY || stateArrayRobot[row][col] == GOAL);
        }
        public void moveRobotToGoal(FloorMap tempFloorMap,int goalRow, int goalColumn)
        {
            
            floorMap = tempFloorMap;
            int row,column;
            while (!isRobotAtGoal(goalRow,goalColumn))
            {
                System.Threading.Thread.Sleep(50);
                row = floorMap.RobotRow;
                column = floorMap.RobotColumn;
                if (row < 0 || column < 0)
                    return;

                if (goalRow == row)
                {   //already at goal row, now move closer to goal column
                    if (!moveRobotToCloserColumn(goalColumn))
                    {
                        eventErrorTryingToRemoveRobot();
                        return;
                    }
                }
                else if(goalColumn == column)
                {   //already at goal column, now move closer to goal row
                    if (!moveRobotToCloserRow(goalRow))
                    {
                        eventErrorTryingToRemoveRobot();
                        return;
                    }
                }
                else
                {   //not at either goal row or goal column, find which is closer
                    if (Math.Abs(goalRow - floorMap.RobotRow) < Math.Abs(goalColumn - floorMap.RobotColumn))
                    {   //closer to goal row than goal column
                        if (!moveRobotToCloserRow(goalRow))
                        { //If it can't move up or down try left or right
                            if (!moveRobotToCloserColumn(goalColumn))
                            {
                                eventErrorTryingToRemoveRobot();
                                return;
                            }
                        }
                    }
                    else
                    {
                        if (!moveRobotToCloserColumn(goalColumn))
                        {   //if it can't move left or right try up or down
                            if (!moveRobotToCloserRow(goalRow))
                            {
                                eventErrorTryingToRemoveRobot();
                                return;
                            }
                                    
                        }
                    }
                }
            }
            eventRobotReacheDestination();
        }
        private void eventErrorTryingToRemoveRobot()
        {
            if (this.errorTryingToMoveRobot != null)
                this.errorTryingToMoveRobot(this, EventArgs.Empty);
        }
        private void eventRobotReacheDestination()
        {
            if (this.robotReachedDestination != null)
                this.robotReachedDestination(this, EventArgs.Empty);
        }
        private bool moveRobotToCloserRow(int goalRow)
        {
            if ((goalRow - floorMap.RobotRow) > 0)
            {
                if (floorMap.moveRobotSouth())
                    return true;
                else
                    return false;
            }
            else
            {
                if (floorMap.moveRobotNorth())
                    return true;
                else
                    return false;
            }
                
        }
        private bool moveRobotToCloserColumn(int goalColumn)
        {
            if ((goalColumn - floorMap.RobotColumn) > 0)
            {
                if (floorMap.moveRobotEast())
                    return true;
                else
                    return false;
            }
            else
            {
                if (floorMap.moveRobotWest())
                    return true;
                else
                    return false;
            }
        }
        private bool isRobotAtGoal(int goalRow, int goalColumn)
        {
            return (floorMap.RobotRow == goalRow && floorMap.RobotColumn == goalColumn);
        }
    }
}
